@@ -10,11 +10,10 @@
 
 #include "lib/jxl/alpha.h"
 #include "lib/jxl/base/byte_order.h"
-#include "lib/jxl/base/padded_bytes.h"
-#include "lib/jxl/base/profiler.h"
-#include "lib/jxl/codec_in_out.h"
-#include "lib/jxl/color_management.h"
+#include "lib/jxl/cms/opsin_params.h"
 #include "lib/jxl/fields.h"
+#include "lib/jxl/frame_header.h"
+#include "lib/jxl/quantizer.h"
 
 namespace jxl {
 BitDepth::BitDepth() { Bundle::Init(this); }
@@ -59,6 +58,7 @@ Status BitDepth::VisitFields(Visitor* JXL_RESTRICT visitor) {
   return true;
 }
 
+#if JXL_DEBUG_V_LEVEL >= 1
 std::string BitDepth::DebugString() const {
   std::ostringstream os;
   os << (floating_point_sample ? "F" : "U");
@@ -66,6 +66,7 @@ std::string BitDepth::DebugString() const {
   if (floating_point_sample) os << "." << exponent_bits_per_sample;
   return os.str();
 }
+#endif
 
 CustomTransformData::CustomTransformData() { Bundle::Init(this); }
 Status CustomTransformData::VisitFields(Visitor* JXL_RESTRICT visitor) {
@@ -252,6 +253,7 @@ Status ExtraChannelInfo::VisitFields(Visitor* JXL_RESTRICT visitor) {
   return true;
 }
 
+#if JXL_DEBUG_V_LEVEL >= 1
 std::string ExtraChannelInfo::DebugString() const {
   std::ostringstream os;
   os << (type == ExtraChannel::kAlpha           ? "Alpha"
@@ -267,6 +269,7 @@ std::string ExtraChannelInfo::DebugString() const {
   os << " shift: " << dim_shift;
   return os.str();
 }
+#endif
 
 ImageMetadata::ImageMetadata() { Bundle::Init(this); }
 Status ImageMetadata::VisitFields(Visitor* JXL_RESTRICT visitor) {
@@ -352,12 +355,13 @@ Status OpsinInverseMatrix::VisitFields(Visitor* JXL_RESTRICT visitor) {
     return true;
   }
   for (int i = 0; i < 9; ++i) {
-    JXL_QUIET_RETURN_IF_ERROR(visitor->F16(
-        DefaultInverseOpsinAbsorbanceMatrix()[i], &inverse_matrix[i]));
+    JXL_QUIET_RETURN_IF_ERROR(
+        visitor->F16(jxl::cms::DefaultInverseOpsinAbsorbanceMatrix()[i],
+                     &inverse_matrix[i]));
   }
   for (int i = 0; i < 3; ++i) {
-    JXL_QUIET_RETURN_IF_ERROR(
-        visitor->F16(kNegOpsinAbsorbanceBiasRGB[i], &opsin_biases[i]));
+    JXL_QUIET_RETURN_IF_ERROR(visitor->F16(
+        jxl::cms::kNegOpsinAbsorbanceBiasRGB[i], &opsin_biases[i]));
   }
   for (int i = 0; i < 4; ++i) {
     JXL_QUIET_RETURN_IF_ERROR(
@@ -401,12 +405,6 @@ Status ReadImageMetadata(BitReader* JXL_RESTRICT reader,
   return Bundle::Read(reader, metadata);
 }
 
-Status WriteImageMetadata(const ImageMetadata& metadata,
-                          BitWriter* JXL_RESTRICT writer, size_t layer,
-                          AuxOut* aux_out) {
-  return Bundle::Write(metadata, writer, layer, aux_out);
-}
-
 void ImageMetadata::SetAlphaBits(uint32_t bits, bool alpha_is_premultiplied) {
   std::vector<ExtraChannelInfo>& eciv = extra_channel_info;
   ExtraChannelInfo* alpha = Find(ExtraChannel::kAlpha);
@@ -443,6 +441,7 @@ void ImageMetadata::SetAlphaBits(uint32_t bits, bool alpha_is_premultiplied) {
   if (bits > 12) modular_16_bit_buffer_sufficient = false;
 }
 
+#if JXL_DEBUG_V_LEVEL >= 1
 std::string ImageMetadata::DebugString() const {
   std::ostringstream os;
   os << bit_depth.DebugString();
@@ -473,5 +472,6 @@ std::string CodecMetadata::DebugString() const {
   os << " " << m.DebugString();
   return os.str();
 }
+#endif
 
 }  // namespace jxl
