@@ -65,7 +65,17 @@ struct State {
   const Header header;
 
   // Allows to approximate division by a number from 1 to 64.
-  uint32_t divlookup[64];
+  //  for (int i = 0; i < 64; i++) divlookup[i] = (1 << 24) / (i + 1);
+
+  const uint32_t divlookup[64] = {
+      16777216, 8388608, 5592405, 4194304, 3355443, 2796202, 2396745, 2097152,
+      1864135,  1677721, 1525201, 1398101, 1290555, 1198372, 1118481, 1048576,
+      986895,   932067,  883011,  838860,  798915,  762600,  729444,  699050,
+      671088,   645277,  621378,  599186,  578524,  559240,  541200,  524288,
+      508400,   493447,  479349,  466033,  453438,  441505,  430185,  419430,
+      409200,   399457,  390167,  381300,  372827,  364722,  356962,  349525,
+      342392,   335544,  328965,  322638,  316551,  310689,  305040,  299593,
+      294337,   289262,  284359,  279620,  275036,  270600,  266305,  262144};
 
   constexpr static pixel_type_w AddBits(pixel_type_w x) {
     return uint64_t(x) << kPredExtraBits;
@@ -78,10 +88,6 @@ struct State {
       pred_errors[i].resize((xsize + 2) * 2);
     }
     error.resize((xsize + 2) * 2);
-    // Initialize division lookup table.
-    for (int i = 0; i < 64; i++) {
-      divlookup[i] = (1 << 24) / (i + 1);
-    }
   }
 
   // Approximates 4+(maxweight<<24)/(x+1), avoiding division
@@ -282,15 +288,15 @@ struct FlatDecisionNode {
     PropertyVal splitval0;
     Predictor predictor;
   };
-  uint32_t childID;  // childID is ctx id if leaf.
   // Property+splitval of the two child nodes.
   union {
     PropertyVal splitvals[2];
     int32_t multiplier;
   };
+  uint32_t childID;  // childID is ctx id if leaf.
   union {
-    int32_t properties[2];
-    int64_t predictor_offset;
+    int16_t properties[2];
+    int32_t predictor_offset;
   };
 };
 using FlatTree = std::vector<FlatDecisionNode>;
@@ -301,22 +307,27 @@ class MATreeLookup {
   struct LookupResult {
     uint32_t context;
     Predictor predictor;
-    int64_t offset;
+    int32_t offset;
     int32_t multiplier;
   };
   JXL_INLINE LookupResult Lookup(const Properties &properties) const {
     uint32_t pos = 0;
     while (true) {
-      const FlatDecisionNode &node = nodes_[pos];
-      if (node.property0 < 0) {
-        return {node.childID, node.predictor, node.predictor_offset,
-                node.multiplier};
-      }
-      bool p0 = properties[node.property0] <= node.splitval0;
-      uint32_t off0 = properties[node.properties[0]] <= node.splitvals[0];
-      uint32_t off1 =
-          2 | (properties[node.properties[1]] <= node.splitvals[1] ? 1 : 0);
-      pos = node.childID + (p0 ? off1 : off0);
+#define TRAVERSE_THE_TREE                                                      \
+  {                                                                            \
+    const FlatDecisionNode &node = nodes_[pos];                                \
+    if (node.property0 < 0) {                                                  \
+      return {node.childID, node.predictor, node.predictor_offset,             \
+              node.multiplier};                                                \
+    }                                                                          \
+    bool p0 = properties[node.property0] <= node.splitval0;                    \
+    uint32_t off0 = properties[node.properties[0]] <= node.splitvals[0];       \
+    uint32_t off1 = 2 | (properties[node.properties[1]] <= node.splitvals[1]); \
+    pos = node.childID + (p0 ? off1 : off0);                                   \
+  }
+
+      TRAVERSE_THE_TREE;
+      TRAVERSE_THE_TREE;
     }
   }
 
@@ -485,8 +496,8 @@ JXL_INLINE PredictionResult Predict(
     // location
     (*p)[offset++] = x;
     // neighbors
-    (*p)[offset++] = std::abs(top);
-    (*p)[offset++] = std::abs(left);
+    (*p)[offset++] = top > 0 ? top : -top;
+    (*p)[offset++] = left > 0 ? left : -left;
     (*p)[offset++] = top;
     (*p)[offset++] = left;
 
@@ -589,6 +600,18 @@ inline PredictionResult PredictTreeWP(Properties *p, size_t w,
       p, w, pp, onerow, x, y, Predictor::Zero, &tree_lookup, &references,
       wp_state, /*predictions=*/nullptr);
 }
+JXL_INLINE PredictionResult PredictTreeWPNEC(Properties *p, size_t w,
+                                             const pixel_type *JXL_RESTRICT pp,
+                                             const intptr_t onerow, const int x,
+                                             const int y,
+                                             const MATreeLookup &tree_lookup,
+                                             const Channel &references,
+                                             weighted::State *wp_state) {
+  return detail::Predict<detail::kUseTree | detail::kUseWP |
+                         detail::kNoEdgeCases>(
+      p, w, pp, onerow, x, y, Predictor::Zero, &tree_lookup, &references,
+      wp_state, /*predictions=*/nullptr);
+}
 
 inline PredictionResult PredictLearn(Properties *p, size_t w,
                                      const pixel_type *JXL_RESTRICT pp,
@@ -612,6 +635,29 @@ inline void PredictLearnAll(Properties *p, size_t w,
       p, w, pp, onerow, x, y, Predictor::Zero,
       /*lookup=*/nullptr, &references, wp_state, predictions);
 }
+inline PredictionResult PredictLearnNEC(Properties *p, size_t w,
+                                        const pixel_type *JXL_RESTRICT pp,
+                                        const intptr_t onerow, const int x,
+                                        const int y, Predictor predictor,
+                                        const Channel &references,
+                                        weighted::State *wp_state) {
+  return detail::Predict<detail::kForceComputeProperties | detail::kUseWP |
+                         detail::kNoEdgeCases>(
+      p, w, pp, onerow, x, y, predictor, /*lookup=*/nullptr, &references,
+      wp_state, /*predictions=*/nullptr);
+}
+
+inline void PredictLearnAllNEC(Properties *p, size_t w,
+                               const pixel_type *JXL_RESTRICT pp,
+                               const intptr_t onerow, const int x, const int y,
+                               const Channel &references,
+                               weighted::State *wp_state,
+                               pixel_type_w *predictions) {
+  detail::Predict<detail::kForceComputeProperties | detail::kUseWP |
+                  detail::kAllPredictions | detail::kNoEdgeCases>(
+      p, w, pp, onerow, x, y, Predictor::Zero,
+      /*lookup=*/nullptr, &references, wp_state, predictions);
+}
 
 inline void PredictAllNoWP(size_t w, const pixel_type *JXL_RESTRICT pp,
                            const intptr_t onerow, const int x, const int y,
