@@ -9,11 +9,12 @@
 
 #include "lib/extras/codec.h"
 #include "lib/jxl/base/data_parallel.h"
+#include "lib/jxl/base/span.h"
 #include "lib/jxl/base/status.h"
-#include "lib/jxl/base/thread_pool_internal.h"
 #include "lib/jxl/codec_in_out.h"
+#include "tools/file_io.h"
+#include "tools/thread_pool_internal.h"
 
-namespace jxl {
 namespace {
 
 // Reads an input file (typically PNM) with color_space hint and writes to an
@@ -27,16 +28,26 @@ int Convert(int argc, char** argv) {
   const std::string& desc = argv[2];
   const std::string& pathname_out = argv[3];
 
-  CodecInOut io;
-  extras::ColorHints color_hints;
-  ThreadPoolInternal pool(4);
+  std::vector<uint8_t> encoded_in;
+  if (!jpegxl::tools::ReadFile(pathname_in, &encoded_in)) {
+    fprintf(stderr, "Failed to read image from %s\n", pathname_in.c_str());
+    return 1;
+  }
+  jxl::CodecInOut io;
+  jxl::extras::ColorHints color_hints;
+  jpegxl::tools::ThreadPoolInternal pool(4);
   color_hints.Add("color_space", desc);
-  if (!SetFromFile(pathname_in, color_hints, &io, &pool)) {
-    fprintf(stderr, "Failed to read %s\n", pathname_in.c_str());
+  if (!jxl::SetFromBytes(jxl::Bytes(encoded_in), color_hints, &io, &pool)) {
+    fprintf(stderr, "Failed to decode %s\n", pathname_in.c_str());
     return 1;
   }
 
-  if (!EncodeToFile(io, pathname_out, &pool)) {
+  std::vector<uint8_t> encoded_out;
+  if (!jxl::Encode(io, pathname_out, &encoded_out, &pool)) {
+    fprintf(stderr, "Failed to encode %s\n", pathname_out.c_str());
+    return 1;
+  }
+  if (!jpegxl::tools::WriteFile(pathname_out, encoded_out)) {
     fprintf(stderr, "Failed to write %s\n", pathname_out.c_str());
     return 1;
   }
@@ -45,6 +56,5 @@ int Convert(int argc, char** argv) {
 }
 
 }  // namespace
-}  // namespace jxl
 
-int main(int argc, char** argv) { return jxl::Convert(argc, argv); }
+int main(int argc, char** argv) { return Convert(argc, argv); }
